#include "bsdisks.h"
#include "adaptors.h"
#include "objectmanager.h"
#include "block.h"
#include "blockfilesystem.h"
#include "drive.h"
#include "filesystemprober.h"
#include "camcontrolprober.h"

DBUSManagerStruct ObjectManager::GetManagedObjects()
{
    return m_managedObjects;
}


void ObjectManager::initialProbe()
{
    GeomProber prober;

    QObject::connect(&prober, &GeomProber::gotDisk,
        [this](DiskInfo di)
        {
            addDrive(di.devName());

            // assume that any object from `geom disk list` has a block device in /dev
            addBlock(di.devName());
        });
    QObject::connect(&prober, &GeomProber::gotPart,
        [this](PartInfo pi)
        {
            foreach(auto devName, pi.partitions().keys())
                addBlock(devName);
        });

    prober.run();
}

void ObjectManager::filesystemAdded(Block* b, QString fs)
{
    Q_ASSERT(b->bFilesystem == nullptr);

    auto bfs = new BlockFilesystem(b);
    bfs->filesystem = fs;

    b->bFilesystem = bfs;
    b->idUsage = QStringLiteral("filesystem");

    if(fs == QStringLiteral("msdosfs"))
    {
        b->idType = QStringLiteral("vfat");
    }
    else
    {
        b->idType = fs;
    }

    addInterface(b->dbusPath, "org.freedesktop.UDisks2.Filesystem", new FilesystemAdaptor(b));
}

Block* ObjectManager::addBlock(QString dev)
{
    // don't do anything with stuff like diskid/blabla
    if(dev.contains('/'))
        return nullptr;

    Q_ASSERT(!m_blockObjects.contains(dev));

    auto* b = new Block;
    m_blockObjects.insert(dev, b);

    QString devPath = UDisksBlockDevices + dev;
    b->dev = ("/dev/" + dev).toLatin1();
    b->dbusPath = QDBusObjectPath(devPath);

    qDebug() << "Registering " + devPath;
    QDBusConnection::systemBus().registerObject(devPath, b);

    startFilesystemProbe(b);
    startGeomProbe(b);

    return b;
}

void ObjectManager::removeBlock(QString dev)
{
    if(m_blockObjects.contains(dev))
    {
        auto* b = m_blockObjects.take(dev);
        qDebug() << "Unregistering " + b->dbusPath.path();
        if(b->bFilesystem)
            removeInterface(b->dbusPath, "org.freedesktop.UDisks2.Filesystem");
        if(b->bPartition)
            removeInterface(b->dbusPath, "org.freedesktop.UDisks2.Partition");
        if(b->bPartTable)
            removeInterface(b->dbusPath, "org.freedesktop.UDisks2.PartitionTable");
        removeInterface(b->dbusPath, "org.freedesktop.UDisks2.Block");
        QDBusConnection::systemBus().unregisterObject(b->dbusPath.path());
        delete b;
    }
    removeDrive(dev);
}

Drive* ObjectManager::addDrive(QString dev)
{
    Q_ASSERT(!m_driveObjects.contains(dev));

    auto* d = new Drive;
    m_driveObjects.insert(dev, d);

    QString devPath = UDisksDrives + dev;
    d->dbusPath = QDBusObjectPath(devPath);
    d->geomName = dev;

    qDebug() << "Registering " + devPath;
    QDBusConnection::systemBus().registerObject(devPath, d);

    auto geomProber = new GeomProber(d->geomName);
    QObject::connect(geomProber, &GeomProber::gotDisk, this,
                     [this, d](DiskInfo di)
    {
        d->size = di.mediaSize();
        d->description = di.descr();
        d->identifier = di.ident();

        d->geomProbeDone = true;
        if(d->camcontrolProbeDone)
            addInterface(d->dbusPath, "org.freedesktop.UDisks2.Drive", new DriveAdaptor(d));
    });

    auto camControlProber = new CamControlProber(d->geomName);
    QObject::connect(camControlProber, &CamControlProber::finished, this,
                     [this, d](bool isRemovable)
    {
        d->isRemovable = isRemovable;
        d->camcontrolProbeDone = true;
        if(d->geomProbeDone)
            addInterface(d->dbusPath, "org.freedesktop.UDisks2.Drive", new DriveAdaptor(d));
    });

    QThreadPool::globalInstance()->start(geomProber);
    QThreadPool::globalInstance()->start(camControlProber);

    return d;
}

void ObjectManager::removeDrive(QString dev)
{
    if(m_driveObjects.contains(dev))
    {
        auto* d = m_driveObjects.take(dev);
        qDebug() << "Unregistering " + d->dbusPath.path();
        removeInterface(d->dbusPath, "org.freedesktop.UDisks2.Drive");
        QDBusConnection::systemBus().unregisterObject(d->dbusPath.path());
        delete d;
    }
}

void ObjectManager::startFilesystemProbe(Block* b)
{
    auto prober = new FilesystemProber(b->dev);

    QObject::connect(prober, &FilesystemProber::finished, this,
                     [this, b](QString fs)
    {
        if(!fs.isEmpty())
            filesystemAdded(b, fs);
    });

    QThreadPool::globalInstance()->start(prober);
}


void ObjectManager::startGeomProbe(Block* b)
{
    auto prober = new GeomProber(QFileInfo(b->dev).fileName());

    QObject::connect(prober, &GeomProber::gotLabels, this,
                     [this, b](QStringList labels)
    {
        b->labels = labels;
    });
    QObject::connect(prober, &GeomProber::gotDisk, this,
                     [this, b](DiskInfo di)
    {
        b->size = di.mediaSize();
        b->description = di.descr();
        b->identifier = di.ident();

        addBlockInterface(b);
    });
    QObject::connect(prober, &GeomProber::gotPart, this,
                     [this, b](PartInfo pi)
    {
        Q_ASSERT(!b->bPartTable);

        b->bPartTable = new BlockPartTable(b);
        b->bPartTable->setTableType(pi.scheme());

        addBlockInterface(b);
        addInterface(b->dbusPath, QStringLiteral("org.freedesktop.UDisks2.PartitionTable"), new PartitionTableAdaptor(b));

        updatePartitions(b, pi.partitions());
    });

    QThreadPool::globalInstance()->start(prober);
}


void ObjectManager::updatePartitions(Block* table, const QHash<QString, Part>& parts)
{
    for(auto it = parts.constBegin(); it != parts.cend(); it++)
    {
        Q_ASSERT(m_blockObjects.contains(it.key()));

        auto* b = m_blockObjects[it.key()];
        auto* bp = new BlockPartition(b);
        auto& pi = it.value();

        bp->partTableBlock = table;
        bp->partitionType = pi.type();
        bp->number = pi.index();
        bp->offset = pi.offset();
        bp->size = pi.length();

        b->bPartition = bp;

        addBlockInterface(b);
        addInterface(b->dbusPath, "org.freedesktop.UDisks2.Partition", new PartitionAdaptor(b));
    }
}

void ObjectManager::addBlockInterface(Block* b)
{
    if(!b->m_initialized)
    {
        b->m_initialized = true;
        addInterface(b->dbusPath, "org.freedesktop.UDisks2.Block", new BlockAdaptor(b));
    }
}


void ObjectManager::addInterface(QDBusObjectPath path, QString iface, QDBusAbstractAdaptor* adaptor)
{
    QVariantMapMap& interfaces = m_managedObjects[path];

    QVariantMap properties;
    for (int i = adaptor->metaObject()->propertyOffset();
            i < adaptor->metaObject()->propertyCount(); ++i) {
        auto propertyName = adaptor->metaObject()->property(i).name();
        properties.insert(QString::fromLatin1(propertyName), adaptor->property(propertyName));
    }

    interfaces[iface] = properties;

    emit InterfacesAdded(path, interfaces);
}


void ObjectManager::removeInterface(QDBusObjectPath path, QString iface)
{
    QVariantMapMap& interfaces = m_managedObjects[path];
    interfaces.remove(iface);
    emit InterfacesRemoved(path, {iface});
}
